/*
 *  linkedlist.c
 *
 *  Description:
 *   Lock-free linkedlist implementation of Harris' algorithm
 *   "A Pragmatic Implementation of Non-Blocking Linked Lists" 
 *   T. Harris, p. 300-314, DISC 2001.
 */

#include "linkedlist.h"

/*
 * The five following functions handle the low-order mark bit that indicates
 * whether a node is logically deleted (1) or not (0).
 *  - is_marked_ref returns whether it is marked, 
 *  - (un)set_marked changes the mark,
 *  - get_(un)marked_ref sets the mark before returning the node.
 */
inline int
is_marked_ref(long i) 
{
  return (int) (i & 0x1L);
}

inline long
unset_mark(long i)
{
  i &= ~0x1L;
  return i;
}

inline long
set_mark(long i) 
{
  i |= 0x1L;
  return i;
}

inline long
get_unmarked_ref(long w) 
{
  return w & ~0x1L;
}

inline long
get_marked_ref(long w) 
{
  return w | 0x1L;
}

/*
 * list_search looks for value val, it
 *  - returns right_node owning val (if present) or its immediately higher 
 *    value present in the list (otherwise) and 
 *  - sets the left_node to the node owning the value immediately lower than val. 
 * Encountered nodes that are marked as logically deleted are physically removed
 * from the list, yet not garbage collected.
 */
node_t* list_search(llist_t* set, val_t val, node_t** left_node) 
{
  return NULL;
}

/*
 * list_contains returns a value different from 0 whether there is a node in the list owning value val.
 */

int list_contains(llist_t* the_list, val_t val)
{
  int result = 1;
  return result;
}


node_t* new_node(val_t val, node_t *next)
{
  node_t* node = NULL;
  return node;
}

llist_t* list_new()
{
  llist_t* the_list = NULL;
  return the_list;
}

void list_delete(llist_t *the_list)
{
}

int list_size(llist_t* the_list)
{
  int size = 0;
  return size;
}

/*
 * list_add inserts a new node with the given value val in the list
 * (if the value was absent) or does nothing (if the value is already present).
 */

int list_add(llist_t *the_list, val_t val)
{
  int result = 1;
  return result;
}

/*
 * list_remove deletes a node with the given value val (if the value is present) 
 * or does nothing (if the value is not present).
 * The deletion is logical and consists of setting the node mark bit to 1.
 */
int list_remove(llist_t *the_list, val_t val)
{
  int result = 1;
  return result;
}

